@@ -53,7 +53,7 @@ generateCommunicationList <- function(model, checkpoint, useSocket, options) {
 				deparse(width.cutoff = 400L, sys.call(-1))), call. = FALSE)
 		}
 		filename <- paste(chkpt.prefix, paste(model$name, 'omx', sep = '.'), sep = '')
-		fullpath <- paste(chkpt.directory, filename, sep="/")
+		fullpath <- file.path(chkpt.directory, filename)
 		override <- mxOption(model, "Checkpoint Fullpath")
 		if (nchar(override)) {
 			fullpath <- override
