@@ -47,8 +47,8 @@
 ##' 
 ##' @details
 ##' x can be the name of an algebra, a bracket address, named entity
-##' or arbitrary expression. It is a frontend-only file that works
-##' much like mxEval. When the \code{details} argument is TRUE, the full
+##' or arbitrary expression.
+##' When the \code{details} argument is TRUE, the full
 ##' sampling covariance matrix of \code{x} is also returned as part of a list.
 ##' The square root of the diagonals of this sampling covariance matrix are
 ##' the standard errors.
@@ -57,6 +57,19 @@
 ##' covariance matrix is given, not the information matrix.  These 
 ##' two are inverses of one another.
 ##' 
+##' This function uses the delta method to compute the standard error of arbitrary
+##' and possibly nonlinear functions of the free parameters.  The delta method
+##' makes a first-order Taylor approximation of the nonlinear function.  The
+##' nonlinear function is a map from all the free parameters to some transformed
+##' subset of parameters: the linearization of this map is given by the Jacobian
+##' \eqn{J}.  In equation form, the delta method computes standard errors by the following:
+##' 
+##' \deqn{J^T C J}
+##' 
+##' where \eqn{J} is the Jacobian of the nonlinear parameter transformation
+##' and \eqn{C} is the covariance matrix of the free parameters (e.g., two
+##' times the inverse of the Hessian of the minus two log likelihood function).
+##' 
 ##' @return SE value(s) returned as a matrix when \code{details} is FALSE.
 ##' When \code{details} is TRUE, a list of the SE value(s) and the full 
 ##' sampling covariance matrix.
@@ -86,36 +99,17 @@
 ##' mxSE(lambda1^2, model = m1)
 mxSE <- function(x, model, details=FALSE, cov, forceName=FALSE, silent=FALSE, ...,
 		 defvar.row=as.integer(NA), data='data'){
-	isCallEtc <- any(c('call', 'language', 'MxAlgebraFormula') %in% is(match.call()$x))
-	if(isCallEtc && !forceName){
-		if(!silent){message('Treating first argument as an expression')}
-		xalg <- mxAlgebraFromString(Reduce(paste, deparse(match.call()$x)), name='onTheFlyAlgebra')
-		x <- "onTheFlyAlgebra"
-		model <- mxModel(model, xalg)
-	} else if('character' %in% is(x)){
-		if(!silent){message('Treating first argument as character named entity in the model')}
-	} else {
-		stop("Please, sir.  'x' must be either the name of an entity in the model, or an expression for an MxAlgebra.")
-	}
-	
-	# Get current algebra/matrix values:
-	freeparams <- omxGetParameters(model)
-	paramnames <- names(freeparams)
-	zoutMat <- try(mxEvalByName(x, model, compute=TRUE),silent=silent)
-	if(class(zoutMat) %in% "try-error"){
-		stop(paste0("I had a problem evaluating your expression. Check that mxEval works for it.\nRecall that elements of submodels are addressed as submodelName.objectName\nFor example, 'sub1.bob' refers to the object 'bob' in the submodel named 'sub1'."))
-	}
-	
 	if(length(model@output) > 0 && missing(cov)){
-		if(length(model@output$infoDefinite) && !single.na(model@output$infoDefinite)){
-			# An indefinite Hessian usually means some SEs will be NaN:
-			ParamsCov <- 2*solve(model@output$hessian)
-			dimnames(ParamsCov) <- dimnames(model@output$hessian)
-		} else {
-			msg <- "Model does not have a reasonable Hessian or standard errors."
-			msg <- paste0(msg, ifelse(imxHasConstraint(model), "\nModel has at least one mxConstraint. This prevented standard error computation.\nTry mxCI().", "\nDid you set the mxOption() to turn off standard errors?"))
+		ParamsCov <- try(vcov(model))
+		if(is(ParamsCov,"try-error")){
+			msg <- "Model does not have a reasonable vcov matrix or standard errors."
 			stop(msg)
 		}
+		# if(length(model@output$infoDefinite) && !single.na(model@output$infoDefinite)){
+		# 	# An indefinite Hessian usually means some SEs will be NaN:
+		# 	ParamsCov <- 2*solve(model@output$hessian)
+		# 	dimnames(ParamsCov) <- dimnames(model@output$hessian)
+		# } else {
 	} else if (missing(cov)){
 		stop("Model does not have output and 'cov' argument is missing.  I'm a doctor, not a bricklayer!\nWas this model run with mxRun?")
 	} else {
@@ -129,13 +123,48 @@ mxSE <- function(x, model, details=FALSE, cov, forceName=FALSE, silent=FALSE, ..
 		}
 	}
 	
+	isCallEtc <- any(c('call', 'language', 'MxAlgebraFormula') %in% is(match.call()$x))
+	ex <- try(eval(x), silent=TRUE)
+	isChar <- !('try-error' %in% is(ex)) && is.character(ex)
+	if(isCallEtc && !forceName && !isChar){
+		if(!silent){message('Treating first argument as an expression')}
+		xalg <- mxAlgebraFromString(Reduce(paste, deparse(match.call()$x)), name='onTheFlyAlgebra')
+		x <- "onTheFlyAlgebra"
+		model <- mxModel(model, xalg)
+	} else if ('character' %in% is(x) && !isCallEtc) {
+		if(!silent){message('Treating first argument as a character')}
+		xalg <- mxAlgebraFromString(Reduce(paste, match.call()$x), name='onTheFlyAlgebra')
+		x <- "onTheFlyAlgebra"
+		model <- mxModel(model, xalg)
+	} else if(isChar){
+		if(!silent){message('Treating first argument as an object that stores a character')}
+		xalg <- mxAlgebraFromString(ex, name='onTheFlyAlgebra')
+		x <- "onTheFlyAlgebra"
+		model <- mxModel(model, xalg)
+	} else {
+		stop("Please, sir.  'x' must be either the name of an entity in the model, or an expression for an MxAlgebra.")
+	}
+	
+	# Get current algebra/matrix values:
+	freeparams <- omxGetParameters(model)
+	paramnames <- names(freeparams)
+	zoutMat <- try(mxEvalByName(x, model, compute=TRUE),silent=silent)
+	if(class(zoutMat) %in% "try-error"){
+		stop(paste0("Couldn't evaluate expression ", omxQuotes(x), ". Might help to check if it works in mxEval.\n",
+		"Recall also that elements of submodels are addressed as submodelName.objectName\n",
+		"For example, to refer to an object called 'bob' in submodel 'sub1', you would say 'sub1.bob'."))
+	}
+	
 	covParam <- ParamsCov
 	jModel <- mxModel(model, mxComputeJacobian(of=x, defvar.row=defvar.row, data=data))
 	jModel <- mxRun(jModel, silent=TRUE)
 	jacTrans <- jModel$compute$output$jacobian
 	covSparam <- jacTrans %*% covParam %*% t(jacTrans)
 	# dimnames(covSparam) <- list(rownames(zoutMat), colnames(zoutMat))
-	SEs <- sqrt(diag(covSparam))
+	if(any(diag(covSparam) < 0) || any(is.na(diag(covSparam)))){
+		warning("Some diagonal elements of the repeated-sampling covariance matrix of the point estimates are less than zero or NA.\nI know, right? Set details=TRUE and check the 'Cov' element of this object.")
+	}
+	SEs <- suppressWarnings(sqrt(diag(covSparam)))
 	SEsMat <- matrix(SEs, nrow = nrow(zoutMat), ncol = ncol(zoutMat))
 	if(details==TRUE){
 		return(list(SE=SEsMat, Cov=covSparam))
